/* Color Cursors 3.0
 * My fourth re-write (Yes, I'm a C-programmer) of my color cursors library.
 *
 * Version History:
 * 3.0 re-writes all routines. Standardizes routine names. Changes file name to Cursors30.c.
 * Adds support for:
 *		multiple cursor regions (Overlapping regions are also defined).
 *		All 4 modifier keys. (with defined precedence, no combinations allowed)
 *		Simple cursors (non-tracking)
 *		Simple animations (aCur-based)
 *		Tracking animations - WOW!
 *
 * 2.1 makes cursor information window-sensitive. Added WindowPtrs to all the calls.
 *
 * 2.0 re-writes all routines. Cursor changing is now more reliable: it verifies
 * where the mouse is with GetMouse, rather than relying on the number of calls to
 * DoMouseMoved. Fixes memory leaks.
 *
 * 1.0. re-write from original verison. Support for special option-key cursors, and
 * a spinning watch cursor; cursor tracking suspension and re-activation. Pulled
 * code into Cursors.c from Eventsx.c
 *
 * 0.0 Original implementation. A crappy hack.
 */

#include "Cursors30.h"
#include "Alertsx.h"

LinkedListPtr gCCWindowList=nil;
WindowPtr	  gCCActiveWindow = nil;
CCCursorPtr	  gCCSimpleCrsr;


RgnHandle CCFindRegion(LinkedListPtr list, RgnHandle rgn);
void CCSetProperRegion(WindowNode *wn);
void CCSetProperCursor(WindowNode *wn, Boolean force);
void CCSetMouseCursor(CCrsrHandle crsr);
void CCBuildUnionRgn(WindowNode *wn);

void DebugRgn(RgnHandle r) ;

/* CCNewCursor
 * The routine creates a new CCCursorRec, initializes the fields in the structure,
 * and returns a handle to the structure. The CCCursorPtr can then be passed into
 * CCSetRegion.
 *
 * If CCNewCursor fails to fully initialize the CCCursorRec, it will return nil.
 */
CCCursorPtr	CCNewCursor(CCType type, short id, Boolean autoDispose)
{
	CCCursorPtr cr;
	short i;
	
	cr = (CCCursorPtr)NewPtr(sizeof(CCCursorRec));
	if (!cr) goto fail;
	
	cr->id = id;
	cr->type = type;
	cr->autoDispose = autoDispose;
	cr->aCur = nil;
	cr->simpleCrsr = nil;
	cr->animCrsrs = nil;
	
	if(type == single) {
		cr->simpleCrsr=GetCCursor(id);
		if (!cr->simpleCrsr) goto fail;
	} else if (type == animation) {
		cr->aCur = (ACURHandle)Get1Resource('acur',id);
		if (!cr->aCur) goto fail;
		
		HLock((Handle)cr->aCur);
		cr->animCrsrs = (CCrsrHandle *)NewPtr(sizeof(CCrsrHandle) * (*(cr->aCur))->numFrames);
		if (!cr->animCrsrs) goto fail;
		for (i=0; i<(*(cr->aCur))->numFrames; i++) cr->animCrsrs[i]=nil;
		
		for (i=0; i<(*(cr->aCur))->numFrames; i++) {
			cr->animCrsrs[i] = GetCCursor((*(cr->aCur))->frames[i*2]);
			if (!cr->animCrsrs[i]) goto fail;
		}
		HUnlock((Handle)cr->aCur);
		cr->animRate = 8;
		cr->changeTicks = 0;
	}
	return cr;
fail:
	if (cr) CCDisposeCursor(cr);
	return nil;
}

/* CCDisposeCursor
 * Used to free memory for a whole or partial CCCursorRec structure. Should be called
 * manually when done with the CCCursorRec. This routine is also called by CCNewCursor
 * if it fails to allocate part of the CCCursorRec structure.
 */
void CCDisposeCursor(CCCursorPtr cr)
{
	short i;

	if (cr) {
		if (cr->simpleCrsr) DisposeCCursor(cr->simpleCrsr);
		if (cr->animCrsrs) {
			for (i=0; i<(*(cr->aCur))->numFrames; i++) {
				if (cr->animCrsrs[i]) DisposeCCursor(cr->animCrsrs[i]);
			}
			DisposePtr((Ptr)cr->animCrsrs);
		}
		if (cr->aCur) ReleaseResource((Handle)cr->aCur);
		DisposePtr((Ptr)cr);
	}	
}

/* CCSetRegion
 * Assigns cursors to a specific region in a specific window.
 */
void CCSetRegion(WindowPtr w, RgnHandle rgn, CCCursorPtr norm, CCCursorPtr cmd, CCCursorPtr opt, CCCursorPtr shift, CCCursorPtr ctl)
{
	RgnHandle rgnKey;
	WindowNode *wn;
	RegionNode *rn;
	short i;

	if (!w) return; if (!rgn) return;
	if (!gCCWindowList) NewLinkedList(&gCCWindowList); // initialize
	
	// find Window in Window List
	wn = (WindowNode*)GetListItemByKey(gCCWindowList,(long)w);
	if (wn == nil) { // our Window isn't in our window list. Add it.
		// make node
		wn = (WindowNode*)NewPtr(sizeof(WindowNode));
		if (!wn) DoError("\pFailed to make WindowNode!",MemError(),kFatal);
		// initialize stuff
		NewLinkedList(&wn->rgns); // make region list
		wn->window = w;
		wn->status = off;
		wn->anim = nil;
		wn->unionRgn = nil;
		// add to window list
		AddItemToList(gCCWindowList, (long)wn, (long)w);
	}
	
	// find Region in Window's Region List
	rgnKey = CCFindRegion(wn->rgns,rgn);
	rn = (RegionNode*)GetListItemByKey(wn->rgns, (long)rgnKey); 
	if (rn == nil) { // you guessed it, add the region to the list
		// make node
		rn = (RegionNode*)NewPtr(sizeof(RegionNode));
		if (!rn) DoError("\pFailed to make RegionNode!",MemError(),kFatal);
		// initialize fields
		for (i=0; i < kCCCursNumber; i++) rn->cursors[i] = nil;
		rn->rgn = rgn;
		AddItemToList(wn->rgns,(long)rn,(long)rgn);// add to window's Rgn list.
		SetPort(w);
		CCBuildUnionRgn(wn); // rebuild our union Rgn.
	}
	
	// set Region data
	rn->cursors[kCCNormalCursor] = norm;
	rn->cursors[kCCCommandCursor] = cmd;
	rn->cursors[kCCOptionCursor] = opt;
	rn->cursors[kCCShiftCursor] = shift;
	rn->cursors[kCCControlCursor] = ctl;
}

void CCRemoveRegion(WindowPtr w, RgnHandle rgn)
{
	WindowNode *wn=nil;
	RegionNode *rn=nil;
	short i;
	RgnHandle  rgnKey;

	if (!w) return; if (!rgn) return; if (!gCCWindowList) return;
	
	// find Window in Window List
	wn = (WindowNode*)GetListItemByKey(gCCWindowList,(long)w);
	if (!wn) return;
	// find Region in Window's Region List. 
	rgnKey = CCFindRegion(wn->rgns, rgn);
	rn = (RegionNode*)GetListItemByKey(wn->rgns,(long)rgnKey);
	if (!rn) return;
	// remove Region from list
	RemoveItemFromList(wn->rgns, (long)rgnKey);
	
	if ((wn->status != off) && (wn->status != simple)) {
		// set proper cursor now that our region is not in the list.
		CCSetProperCursor(wn, true);
	}
	
	for (i=0; i < kCCCursNumber; i++) {
		if (rn->cursors[i]->autoDispose) {
			CCDisposeCursor(rn->cursors[i]);
			rn->cursors[i] = nil;
		}
	}
	DisposePtr((Ptr)rn);
	
	CCBuildUnionRgn(wn); // rebuild our union Rgn.
}

void CCRemoveWindow(WindowPtr w)
{
	WindowNode *wn=nil;
	RegionNode *rn=nil;
	short i,j;
	RgnHandle  rgn;

	if (!w) return; if (!gCCWindowList) return;
	
	// find Window in Window List
	wn = (WindowNode*)GetListItemByKey(gCCWindowList,(long)w);
	if (!wn) return;
	
	if (gCCActiveWindow == w) gCCActiveWindow = nil;
	
	// find Region in Window's Region List.
	for (i=1; i<=LinkedListCount(wn->rgns); i++) {
		rn = (RegionNode*)GetListItemByRank(wn->rgns, i);
		// wipe Region out completely.
		RemoveItemFromList(wn->rgns, (long)rn->rgn);
		DisposeRgn(rn->rgn);
		for (j=0; j < kCCCursNumber; j++) {
			if (rn->cursors[j]->autoDispose) {
				CCDisposeCursor(rn->cursors[j]);
				rn->cursors[j] = nil;
			}
		}
		
		DisposePtr((Ptr)rn);
	}
	
	DisposeLinkedList(wn->rgns);
	DisposeRgn(wn->unionRgn);
	RemoveItemFromList(gCCWindowList,(long)w);
	DisposePtr((Ptr)wn);
}


void CCTrackCursor(WindowPtr w, Boolean enable)
{
	WindowNode *wn=nil;
	RegionNode *rn=nil;
	short i;
	RgnHandle  rgnKey;

	if (!w) return; if (!gCCWindowList) return;
	
	// find Window in Window List
	wn = (WindowNode*)GetListItemByKey(gCCWindowList,(long)w); if (!wn) return;
	
	if (enable) {
		wn->status = norml;
		gCCActiveWindow = w;
		
		SetPort(w);
		CCSetProperRegion(wn);
		CCSetProperCursor(wn,true);
	}else{
		if (gCCActiveWindow == w) { // double check.
			wn->status = off;
			gCCActiveWindow = nil;
			SetCursor(&qd.arrow); // set the default cursor.
		}
	}
}

/* CCMouseMoved
 * Called from the event loop when program receives a mouse-moved event. This event
 * *REALLY* is an osEvt. Read up on it ... 
 */
void CCMouseMoved(void)
{
	WindowNode *wn=nil;

	if (!gCCWindowList) return ;
	if (!gCCActiveWindow) return ;
	wn = (WindowNode*)GetListItemByKey(gCCWindowList,(long)gCCActiveWindow); 
	if (!wn) return;
	
	SetPort(gCCActiveWindow);
	CCSetProperRegion(wn);
	if (wn->status != simple) CCSetProperCursor(wn,true);
}

/* CCEventProcessing 
 * Called periodically from event loop so the cursor will respond to modifier
 * keys.
 */
void CCEventProcessing(void)
{
	WindowNode *wn=nil;

	if (!gCCWindowList) return;
	if (!gCCActiveWindow) return;
	wn = (WindowNode*)GetListItemByKey(gCCWindowList,(long)gCCActiveWindow); 
	if (!wn) return;
	if (wn->status == simple) return;

	CCSetProperCursor(wn,false);
	
	if (wn->anim) {
		if (wn->anim->changeTicks < TickCount()) {
			CCAnimateCursor(wn);
			wn->anim->changeTicks = wn->anim->animRate + TickCount();
		}
	}
}

/* CCMouseRgn
 * Returns a RgnHandle to be used with WaitNextEvent.
 */
RgnHandle CCMouseRgn(void)
{
	WindowNode *wn=nil;
	RgnHandle ret;

	if (!gCCWindowList) return nil;
	if (!gCCActiveWindow) return nil;
	
	wn = (WindowNode*)GetListItemByKey(gCCWindowList, (long)gCCActiveWindow);
	if (!wn) return nil;
	
	if (wn->activeRN) ret= wn->activeRN->rgn;
	else ret= wn->unionRgn;
	
	return ret;
}

void CCSimpleCursor(CCType type, short crsrID, Boolean enable)
{
	WindowNode *wn=nil;

//	if ((gCCActiveWindow) && (gCCWindowList))
//		wn = (WindowNode*)GetListItemByKey(gCCWindowList,(long)gCCActiveWindow); 

	if (enable) {
		if (gCCSimpleCrsr) CCDisposeCursor(gCCSimpleCrsr);
		gCCSimpleCrsr = CCNewCursor(type, crsrID, true);
		if (gCCSimpleCrsr) {
			if (type == single) {
				CCSetMouseCursor(gCCSimpleCrsr->simpleCrsr);
			} else if(type == animation) {
				CCSetMouseCursor(gCCSimpleCrsr->animCrsrs[(*(gCCSimpleCrsr->aCur))->curFrame]);
			}
		}
	//	if (wn) wn->status = simple;
	} else {
		if (gCCSimpleCrsr) CCDisposeCursor(gCCSimpleCrsr);
		gCCSimpleCrsr = nil;
		if ((gCCActiveWindow) && (gCCWindowList)) 
			wn = (WindowNode*)GetListItemByKey(gCCWindowList,(long)gCCActiveWindow);
		CCSetProperCursor(wn,true); // wn can be nil here
	}
}

/* CCAnimateCursor
 * Cause the current animated cursor to advance one frame. The active cursor
 * must be an animated cursor (though it is safe to call the routine if the 
 * active cursor is single). To call this routine manually, wn must be set
 * to nil; otherwise, wn is a pointer to the WindowNode for the frontmost window.
 */ 
/* Probably ought to use CursorCtl.h - MDT 3/27/96
 */
void CCAnimateCursor(WindowNode *wn)
{
	CCCursorPtr cr=nil;
	
	if (!wn) { // simple (manual)
		cr = gCCSimpleCrsr;
	} else {
		cr = wn->anim;
	}
	if (!cr) return;
	if (cr->type != animation) return;
	
	CCSetMouseCursor(cr->animCrsrs[(*(cr->aCur))->curFrame++]);
	if ((*(cr->aCur))->curFrame >= (*(cr->aCur))->numFrames)
		(*(cr->aCur))->curFrame = 0;
}


/**
 ** PRIVATE
 **/

/* CCFindRegion
 * Locates Region rgn in List list. Used because two different RgnHandles can 
 * refer to the same region. If rgn is found, CCFindRegion will return the RgnHandle
 * used to identify the RegionNode item in list. If rgn is not in list, CCFindRegion 
 * returns nil.
 */
RgnHandle CCFindRegion(LinkedListPtr list, RgnHandle rgn)
{
	RgnHandle rgnKey;
	short i;
	
	for (i=1; i<=LinkedListCount(list); i++) {
		rgnKey = (RgnHandle)GetListKeyByRank(list, i);
		if (rgn == rgnKey) return rgnKey; // optomization: if handles are identical, their regions are identical.
		if (EqualRgn(rgn, rgnKey)) return rgnKey;
	}

	return nil;
}

/* CCSetProperRegion
 * Dig through a Window's region. Find out which region wins, and set the active 
 * region. Assumes port is set.
 */
void CCSetProperRegion(WindowNode *wn)
{
	RegionNode *rn=nil;
	short i;
	Point	mouse;

	if (!wn) return;
	GetMouse(&mouse); // need port set
	LocalToGlobal(&mouse);
	
	wn->activeRN = nil;
	for (i=1; i<=LinkedListCount(wn->rgns); i++) {
		rn = (RegionNode*)GetListItemByRank(wn->rgns, i);
		if (PtInRgn(mouse,rn->rgn)) {
			wn->activeRN = rn;
			break;
		}
	}
}

#define kCmdKey   0x0037
#define kShiftKey 0x0038
#define kOptKey	  0x003A
#define kCtlKey   0x003B

/* CCSetProperCursor
 * Sets cursor appropriately based on what modifier key is down. ONLY CALLED WHEN TRACKING
 */
void CCSetProperCursor(WindowNode *wn, Boolean force)
{
	RegionNode *rn=nil;
	unsigned char km[16];
	CCStatus	origStatus;
	CCCursorPtr cr;
	
	if (wn) 
		rn = wn->activeRN;
	if (!rn) {
		CCSetMouseCursor(nil);
		if (wn) {
			wn->status = arrow;
			wn->anim = nil;
		}
		return;
	}
	origStatus = wn->status;
	
	GetKeys( (unsigned long *) km);
	
	if ( ( km[kCmdKey>>3] >> (kCmdKey & 7) ) & 1) {				// command
		wn->status = command;
		cr = rn->cursors[kCCCommandCursor];
	} else if ( ( km[kOptKey>>3] >> (kOptKey & 7) ) & 1) {		// option
		wn->status = option;
		cr = rn->cursors[kCCOptionCursor];
	} else if ( ( km[kShiftKey>>3] >> (kShiftKey & 7) ) & 1) {	// shift
		wn->status = shift;
		cr = rn->cursors[kCCShiftCursor];
	} else if ( ( km[kCtlKey>>3] >> (kCtlKey & 7) ) & 1) {		// control
		wn->status = control;
		cr = rn->cursors[kCCControlCursor];
	} else if ((force) || (wn->status != norml)) {				// normal
		wn->status=norml;
		cr = rn->cursors[kCCNormalCursor];
	}
	
	if ((force) || (origStatus != wn->status)) {
		if (!cr) cr = rn->cursors[kCCNormalCursor]; // if no modifier cursor, use normal
		if (!cr) {
			CCSetMouseCursor(nil);               // if no normal, set arrow
		} else {
			if(cr->type == single) {
				CCSetMouseCursor(cr->simpleCrsr);
				wn->anim = nil;
			} else if(cr->type == animation) {
				CCSetMouseCursor(cr->animCrsrs[(*(cr->aCur))->curFrame]);
				wn->anim = cr;
			}
		}
	}
}

void CCSetMouseCursor(CCrsrHandle crsr)
{
	if (crsr == nil) SetCursor(&qd.arrow);
	else SetCCursor(crsr);
}

/*
 * Assumes port is set!
 */
void CCBuildUnionRgn(WindowNode *wn)
{
	short i;
	RegionNode *rn;
	RgnHandle tmp = NewRgn();
	
	if (wn->unionRgn) DisposeRgn(wn->unionRgn);
	wn->unionRgn = NewRgn();
	for (i=1; i<=LinkedListCount(wn->rgns); i++) {
		rn = (RegionNode*)GetListItemByRank(wn->rgns, i);
		UnionRgn(tmp,rn->rgn, tmp);
	}
	OpenRgn();
	FrameRect(&qd.screenBits.bounds); // not multimonitor friendly!
	FrameRgn(tmp);
	CloseRgn(wn->unionRgn);
	DisposeRgn(tmp);
}