/*
 * MultiPane.c
 *
 * Routines for handling multi-pane dialog boxes. Design loosely based on some code I 
 * saw in Develop 23.
 *
 * MDT 3/11/96
 */

#include "MultiPane.h"

#ifdef __PERFORM_ITEM_CACHING__
short InitMPDialogItems(MPDialogRecord  *mpDRec);
void SaveMPDialogItems(MPDialogRecord  *mpDRec);
void RestoreMPDialogItems(MPDialogRecord  *mpDRec);
void DisposeMPDialogItems(MPDialogRecord  *mpDRec);
#endif // __PERFORM_ITEM_CACHING__

/* GetNewMPDialog
 * Gets a new MPDialog from the resource fork. An MPDialog is composed of the following 
 * resources: A DLOG of id dID, a dtl# of id dID. The DLOG is the master dialog box; 
 * the dialog that contains all items common to our sub-views. The dtl# contains a list
 * of DITLs composing our sub-panes. The format of the dtl# resource is as follows:
 * 		0 - 1	Number of entries (2 bytes)
 *		2 - 3	Entry number 1 (2 bytes)
 *		etc.
 * Sub-pane items will be overlaid atop the master dialog.
 */
DialogPtr GetNewMPDialog(short dID, Ptr dStorage, WindowPtr behind)
{
	MPDialogRecord  *mpDRec;
	Handle			handle;
	short			cnt,i,j;
	
	//evil
	/*ItemListHdl		ditl;
	DITLItemPtr		di;*/
	
	if (!dStorage) {
		mpDRec = (MPDialogRecord *) NewPtr(sizeof(MPDialogRecord));
	}
	if (!mpDRec) return nil;

	// get master dialog info
	mpDRec->dlogPtr = GetNewDialog(dID,&mpDRec->dlogRecord,behind);
	if (!mpDRec->dlogPtr) return nil;
	mpDRec->dlogID = dID;
	mpDRec->dlogItemCnt = CountDITL(mpDRec->dlogPtr);

	// get sub-pane info
	handle = Get1Resource('dtl#',dID);
	if (!handle) return nil;
	cnt = ((short*)*handle)[0];
	mpDRec->subPaneIDs = (short*)NewPtr(sizeof(short) * cnt);
	for (i=0; i<cnt; i++) {
		mpDRec->subPaneIDs[i] = ((short*)*handle)[i+1];
	}
	ReleaseResource(handle);
	mpDRec->numPanes = cnt;
	mpDRec->subPaneNum  = -1;
	mpDRec->customErase = -1;
	
#ifdef __PERFORM_ITEM_CACHING__
	// evil dabbling items - record item types
	if (InitMPDialogItems(mpDRec)) return nil;
#endif // __PERFORM_ITEM_CACHING__
	
	// return dialog ID.
	return (mpDRec->dlogPtr);
}

/* CloseMPDialog
 * This removes the Dialog resource and most memory w/o removing the MPDialogRecord.
 * Just what you would expect.
 */
void CloseMPDialog(DialogPtr mpDialog)
{
	if (!mpDialog) return;
	
	CloseDialog(mpDialog);
}

/* DisposeMPDialog
 * this removes the Dialog resource, and all memory associated ith it. Just what you would
 * expect it to do.
 */
void DisposeMPDialog(DialogPtr mpDialog)
{
	MPDialogRecord  *mpDRec;
	DialogPeek		dp;
	
	if (!mpDialog) return;
	
	// get MPRec Info.
	dp = (DialogPeek)mpDialog;
	mpDRec = (MPDialogRecord *) dp;
	
	// close the dialog
	CloseDialog(mpDRec->dlogPtr);
	
	// dispose the MPDialogRecord.
#ifdef __PERFORM_ITEM_CACHING__
	DisposeMPDialogItems(mpDRec);
#endif // __PERFORM_ITEM_CACHING__
	DisposePtr((Ptr)mpDRec->subPaneIDs);
	DisposePtr((Ptr)mpDRec);
}

/* SetMPDialogPane
 * This routine is used to switch between panes on a dialog. First, it removes any extra
 * items added by previous calls to SetMPDialogPane. Second, it erases the area specified
 * by an optional customErase dialog item. Third, it then builds the new pane.
 * paneNum is the pane number to add; it must be between 0 and the number of panes-1. If
 * paneNum is set to -1, the master dialog is returned to its original state.
 */
void SetMPDialogPane(DialogPtr mpDialog, short paneNum)
{
	MPDialogRecord  *mpDRec;
	DialogPeek		dp;
	Handle 			handle;
	short			i;
		
	if (!mpDialog) return;
	
	// get MPRec Info.
	dp = (DialogPeek)mpDialog;
	mpDRec = (MPDialogRecord *) dp;
	
	// ditch early if we don't really need to switch.
	if (mpDRec->subPaneNum == paneNum) return;
	
#ifdef __PERFORM_ITEM_CACHING__
	// Evil Save
	SaveMPDialogItems(mpDRec);
#endif // __PERFORM_ITEM_CACHING__

	// set new paneNum
	mpDRec->subPaneNum = paneNum;
	
	// remove excess D items.
	i = CountDITL(mpDRec->dlogPtr) - mpDRec->dlogItemCnt;
	ShortenDITL(mpDRec->dlogPtr,i);
	
	// optional customErase Rect MDT 5/13/96
	if (mpDRec->customErase > 0) {
		short iType;
		Rect iRect;
		Handle iHandle;
		GetDItem(mpDialog, mpDRec->customErase, &iType, &iHandle, &iRect);
		if (iRect.left > 16380) { // rounded down from 16384 -- just in case.
			OffsetRect(&iRect,-16384,0); 
		}
		EraseRect(&iRect);
	}
	
	// check paneNum; do we bail?
	if ((paneNum < 0) || (paneNum >= mpDRec->numPanes)) return;

	// get DITL info
	handle = Get1Resource('DITL',mpDRec->subPaneIDs[paneNum]);
	if (!handle) return;
	
	// append & cleanup.
	AppendDITL(mpDRec->dlogPtr, handle, overlayDITL);
	ReleaseResource(handle);

#ifdef __PERFORM_ITEM_CACHING__
	// Evil Restore
	RestoreMPDialogItems(mpDRec);
#endif // __PERFORM_ITEM_CACHING__
}

/* GetMPDialogPane
 * This routine returns the active pane number.
 */
short GetMPDialogPane(DialogPtr mpDialog)
{
	MPDialogRecord  *mpDRec;
	DialogPeek		dp;
	
	if (!mpDialog) return;
	
	// get MPRec Info.
	dp = (DialogPeek)mpDialog;
	mpDRec = (MPDialogRecord *) dp;
	
	return mpDRec->subPaneNum;
}

/* MPItemHit
 * Converts the itemHit returned by ModalDialog into local subPane item numbers and
 * returns the local subPane number.
 */
void MPItemHit(DialogPtr mpDialog, short *itemHit, short *paneNum)
{
	MPDialogRecord  *mpDRec;
	DialogPeek		dp;
	
	if (!mpDialog) return;
	
	// get MPRec Info.
	dp = (DialogPeek)mpDialog;
	mpDRec = (MPDialogRecord *) dp;
	
	if (*itemHit <= mpDRec->dlogItemCnt) {
		//item is in our default pane. Leave it be.
		*paneNum = -1;
	} else {
		*itemHit -= mpDRec->dlogItemCnt;
		*paneNum = mpDRec->subPaneNum;
	}
}

/* GetMPDialogItemCount
 * Returns the number of items in the main pane. Subtract it from the item number
 * to get the item number local to the current subpane. Very much like programming
 * a cdev.
 */
short GetMPDialogItemCount(DialogPtr mpDialog)
{
	MPDialogRecord  *mpDRec;
	DialogPeek		dp;
	
	if (!mpDialog) return 0;
	
	// get MPRec Info.
	dp = (DialogPeek)mpDialog;
	mpDRec = (MPDialogRecord *) dp;
	
	return mpDRec->dlogItemCnt;
}

/* SetMPCustomEraseItem - OPTIONAL
 * Identify an area of the Dialog to erase every time the active Pane is switched.
 * Typically, this is a user item on the main dialog that blankets the region the
 * sub Panes will appear in. Note that the Dialog Manager will clean up after itself;
 * you will proabably only want to use this if you're doing some graphics work in
 * a sub-pane (i.e., Drawing a box around some radio buttons).
 */
void SetMPCustomEraseItem(DialogPtr mpDialog, short item)
{
	MPDialogRecord  *mpDRec;
	DialogPeek		dp;
	
	if (!mpDialog) return;
	dp = (DialogPeek)mpDialog;
	mpDRec = (MPDialogRecord *) dp;
	mpDRec->customErase = item;
	
	// Now hide the item so it doesn't interfere w/ our drawing.
	HideDItem(mpDialog, item);
}

#ifdef __PERFORM_ITEM_CACHING__

void SetMPCtlValue(DialogPtr mpDialog, short value, short item, short pane)
{
	MPDialogRecord  *mpDRec;
	DialogPeek		dp;
	short iType;
	Handle iHandle;
	Rect iRect;
	short itm = item-1; //item - 1;
	
	if (!mpDialog) return;
	dp = (DialogPeek)mpDialog;
	mpDRec = (MPDialogRecord *) dp;

	// set value in our private data area
	if ((pane < -1) || (pane >= mpDRec->numPanes)) return; // pane out of range
	if (pane > -1) {
		if ((itm < 0) || (itm >= mpDRec->subPaneItems[pane].itemCnt)) return; // item out of range
		if (!((mpDRec->subPaneItems[pane].mpDITLItems[itm].itmType & 0x03) >0)) return;
		mpDRec->subPaneItems[pane].mpDITLItems[itm].itmValue = value;
	}
	
	// if the item is on screen, the set the item there too.
	if ((mpDRec->subPaneNum == pane) || (pane == -1)) {
		if (pane != -1) item += mpDRec->dlogItemCnt;
		GetDItem(mpDialog,item,&iType,&iHandle,&iRect); 
		SetCtlValue((ControlHandle)iHandle,value);
	}
}

short GetMPCtlValue(DialogPtr mpDialog, short item, short pane)
{
	MPDialogRecord  *mpDRec;
	DialogPeek		dp;
	short itm = item -1,ret;
	short iType;
	Handle iHandle;
	Rect iRect;
	
	if (!mpDialog) return;
	dp = (DialogPeek)mpDialog;
	mpDRec = (MPDialogRecord *) dp;
	
	if ((pane < -1) || (pane >= mpDRec->numPanes)) return; // pane out of range
	if (pane != -1) {
		if ((itm < 0) || (itm >= mpDRec->subPaneItems[pane].itemCnt)) return; // item out of range
		if (!((mpDRec->subPaneItems[pane].mpDITLItems[itm].itmType & 0x03) >0)) return 0;
	}
	
	// if our control is on screen, we should read it's value directly -- it hasn't been
	// filed into our database yet.
	if ((mpDRec->subPaneNum == pane) || (pane == -1)) {
		if (pane != -1) item += mpDRec->dlogItemCnt;
		GetDItem(mpDialog,item,&iType,&iHandle,&iRect); 
		ret = GetCtlValue((ControlHandle)iHandle);
	} else { // otherwise, pull it from our database.
		ret = mpDRec->subPaneItems[pane].mpDITLItems[itm].itmValue;
	}
	
	return ret;
}

void SetMPIText(DialogPtr mpDialog,Str255 str, short item, short pane)
{
	MPDialogRecord  *mpDRec;
	DialogPeek		dp;
	short iType;
	Handle iHandle;
	Rect iRect;
	short itm = item-1; //item - 1;
	
	if (!mpDialog) return;
	dp = (DialogPeek)mpDialog;
	mpDRec = (MPDialogRecord *) dp;
	
	if ((pane < -1) || (pane >= mpDRec->numPanes)) return; // pane out of range
	if (pane > -1) {
		if ((itm < 0) || (itm >= mpDRec->subPaneItems[pane].itemCnt)) return; // item out of range
		if (!(mpDRec->subPaneItems[pane].mpDITLItems[itm].itmType & 0x10)) return;
		BlockMove(str,
				  mpDRec->subPaneItems[pane].mpDITLItems[itm].itmString,
				  str[0]+1);
	}
			  
	if ((mpDRec->subPaneNum == pane) || (pane == -1)) {
		if (pane != -1) item += mpDRec->dlogItemCnt;
		GetDItem(mpDialog,item,&iType,&iHandle,&iRect);
		SetIText(iHandle,str);
	}
}

void GetMPIText(DialogPtr mpDialog,Str255 str, short item, short pane)
{
	MPDialogRecord  *mpDRec;
	DialogPeek		dp;
	short iType;
	Handle iHandle;
	Rect iRect;
	short itm = item - 1;
	
	if (!mpDialog) return;
	dp = (DialogPeek)mpDialog;
	mpDRec = (MPDialogRecord *) dp;

	if ((pane < -1) || (pane >= mpDRec->numPanes)) return; // pane out of range
	if (pane != -1) {
		if ((itm < 0) || (itm >= mpDRec->subPaneItems[pane].itemCnt)) return; // item out of range
		if (!(mpDRec->subPaneItems[pane].mpDITLItems[itm].itmType & 0x10)) return;
	}
	
	// if on screen, use it
	if ((mpDRec->subPaneNum == pane) || (pane == -1)) {
		if (pane != -1) item += mpDRec->dlogItemCnt;
		GetDItem(mpDialog,item,&iType,&iHandle,&iRect);
		GetIText(iHandle, str);
	} else {
		BlockMove(mpDRec->subPaneItems[pane].mpDITLItems[itm].itmString,
				  str,
				  mpDRec->subPaneItems[pane].mpDITLItems[itm].itmString[0]+1);
	}
}

#endif // __PERFORM_ITEM_CACHING__


/*
 * Private Functions
 */

#ifdef __PERFORM_ITEM_CACHING__

/* InitMPDialogItems
 * This routine loops through all the items in all the DITLs specified in a Dialog's
 * 'dtl#' resource, and creates some special datatypes used in item caching. This is
 * only called at the end of GetNewMPDialog.
 */
short InitMPDialogItems(MPDialogRecord  *mpDRec)
{
	ItemListHdl		ditl;
	DITLItemPtr		di;
	short i,j, ditlCnt, itemCnt;
	
	if (!mpDRec) return -1;
	
	// make a new MPItemList Array
	ditlCnt = mpDRec->numPanes;
	mpDRec->subPaneItems = (MPItemListPtr)NewPtr(sizeof(MPItemList) * ditlCnt);
	if (!mpDRec->subPaneItems) return -1;
	
	// loop through all ditls and fill in our MPItemList array
	for (i=0; i<ditlCnt; i++) {
		ditl = (ItemListHdl) Get1Resource('DITL',mpDRec->subPaneIDs[i]);
		if (!ditl) return nil;
		
		// for each ditl, loop through every item. Because the structure of the DITL
		// resource is dynamic, we'll use a di ptr and adjust it's address manually;
		// with variable record sizes, and with padding to keep longs on word bound-
		// aries, we can't trust ptr arithmetic.
		mpDRec->subPaneItems[i].itemCnt = (*ditl)->dlgMaxIndex;
		mpDRec->subPaneItems[i].mpDITLItems = 
						(MPDITLItemPtr) NewPtr(sizeof(MPDITLItem) * (*ditl)->dlgMaxIndex) ;
		if (!mpDRec->subPaneItems[i].mpDITLItems) return -1;
		di = (*ditl)->DITLItems; // again, di points to the dialog item we are currently working on.
		for (j=0; j < (*ditl)->dlgMaxIndex; j++) {
			// for each item, record it's type. If it's any button but push button, 
			// set a default value for the control's value. If it's an edit text field, 
			// store its initial value.
			mpDRec->subPaneItems[i].mpDITLItems[j].itmType = di->itmType;
			if ((di->itmType & 0x03) > 0) { // if it's a control
				mpDRec->subPaneItems[i].mpDITLItems[j].itmValue = 0; 
			}
			if (di->itmType & 0x10) { // if it's a string
				BlockMove(di->itmData,
						mpDRec->subPaneItems[i].mpDITLItems[j].itmString,
						di->itmData[0]+1);
			}
			
			// here we need to adjust our Dialog Item pointer by hand. Note that 
			// the final address must be even, ensuring that the record starts on
			// a word boundary.
			di = (DITLItemPtr)(((long)di) +14 + di->itmData[0] );
			if (((long)di) & 0x00000001 == 1) di = (DITLItemPtr)(((long)di) + 1); // if odd.
		}
	}
	
	return 0;
}

/* SaveMPDialogItems
 * Save our subPane items into our evil data structure.
 */
void SaveMPDialogItems(MPDialogRecord  *mpDRec)
{
	short item, itemCnt;
	short masterItems, pane,totalItems;
	short iType;
	Handle iHandle;
	Rect iRect;
	Str255 str;
	
	if (!mpDRec) return;
	
	masterItems = mpDRec->dlogItemCnt;
	pane = mpDRec->subPaneNum;
	totalItems = CountDITL(mpDRec->dlogPtr);
	
	for (item = masterItems; item <totalItems; item++) {
		GetDItem(mpDRec->dlogPtr,item+1,&iType,&iHandle,&iRect);
		if (iHandle == nil) continue;
		if ((iType & 0x03) > 0) { // if it's a control
			mpDRec->subPaneItems[pane].mpDITLItems[item-masterItems].itmValue = 
											GetCtlValue((ControlHandle)iHandle);
		}
		if (iType & 0x10) { // if it's a string
			GetIText(iHandle,str);
			BlockMove(str,
					mpDRec->subPaneItems[pane].mpDITLItems[item-masterItems].itmString,
					str[0]+1);
		}
	} 
}

/* RestoreMPDialogItems
 * Save our subPane items into our evil data structure. See SaveMPDialogItems ...
 */
void RestoreMPDialogItems(MPDialogRecord  *mpDRec)
{
	short item, itemCnt;
	short masterItems, pane,totalItems;
	short iType;
	Handle iHandle;
	Rect iRect;
	Str255 str;
	
	if (!mpDRec) return;
	
	masterItems = mpDRec->dlogItemCnt;
	pane = mpDRec->subPaneNum;
	totalItems = CountDITL(mpDRec->dlogPtr);
	
	for (item = masterItems; item <totalItems; item++) {
		GetDItem(mpDRec->dlogPtr,item+1,&iType,&iHandle,&iRect);
		if (iHandle == nil) continue;
		if ((iType & 0x03) > 0) { // if it's a control
			SetCtlValue((ControlHandle)iHandle,
					mpDRec->subPaneItems[pane].mpDITLItems[item-masterItems].itmValue);
		}
		if (iType & 0x10) { // if it's a string
			BlockMove(mpDRec->subPaneItems[pane].mpDITLItems[item-masterItems].itmString,
					str,
					mpDRec->subPaneItems[pane].mpDITLItems[item-masterItems].itmString[0]+1);
			SetIText(iHandle,str);
		}
	} 
}

void DisposeMPDialogItems(MPDialogRecord  *mpDRec)
{
	short i;

	if (mpDRec->subPaneItems) {
		for (i=0; i<mpDRec->numPanes; i++) {
			if (mpDRec->subPaneItems[i].mpDITLItems) 
				DisposePtr((Ptr)mpDRec->subPaneItems[i].mpDITLItems);
		}
		
		DisposePtr((Ptr)mpDRec->subPaneItems);
	}
}

#endif // __PERFORM_ITEM_CACHING__