/*
 * WindowUtil.c
 * 
 * Some common utilities for dealing with windows.
 */

#include "WindowUtil.h"

// Tiling constants:
#define kGapBetweenWindows 2

// Stacking constants:
#define kVerticalStagger	18
#define kHorizontalStagger	5
#define kWindowsPerDiagonal	8


void SetWindowBounds(WindowPtr w, Rect *r);
void FEWTileWindows(WindowPtr w, Ptr d);
void FEWStackWindows(WindowPtr w, Ptr d);
void FEWCountWindows(WindowPtr w, Ptr data);
void ForEachWindowDo(void (*func)(WindowPtr wp, Ptr d), Ptr data, Boolean (*exclude)(WindowPtr wpp));

/* StackWindows
 */
void StackWindows(void (*autosize)(WindowPtr ww), Boolean (*func)(WindowPtr wp))
{
	short screenWidth, screenHeight;
	short maxWindowsInDiagonal;
	short numberOfBottomWindows;
	TileStackWindowData data;

	//count windows
	data.currentWindow = data.numberOfWindows = 0; 
	ForEachWindowDo(FEWCountWindows,(Ptr)&data.numberOfWindows,func);

	data.screenRect = qd.screenBits.bounds;
	data.screenRect.top += GetMBarHeight();
	InsetRect(&data.screenRect, kGapBetweenWindows, kGapBetweenWindows);
	
	screenWidth = data.screenRect.right - data.screenRect.left;
	screenHeight = data.screenRect.bottom - data.screenRect.top;
	
	maxWindowsInDiagonal = data.numberOfWindows;
	if (maxWindowsInDiagonal > kWindowsPerDiagonal) maxWindowsInDiagonal = kWindowsPerDiagonal;
	
	numberOfBottomWindows = data.numberOfWindows / kWindowsPerDiagonal;
	if (numberOfBottomWindows > 0) --numberOfBottomWindows;
	
	data.width = screenWidth - (((maxWindowsInDiagonal - 1) + numberOfBottomWindows) * kHorizontalStagger);
	data.height= screenHeight - (maxWindowsInDiagonal - 1) * kVerticalStagger;// * kHorizontalStagger;
	
	data.autosize = autosize;
	ForEachWindowDo(FEWStackWindows,(Ptr)&data,func);
}

/* TileWindows
 * Tiles all windows on the screen. There is an optional descrimination function that 
 * allows you to exclude certain windows:
 *    Boolean Exclusion(WindowPtr w)
 * It should examine the WindowPtr and return true if the window should be excluded
 * from the Tile operation.
 *
 * Routine loosely modeled after Scott Knaster.
 */
void TileWindows(short horv, short minWidth, /*short minHeight,*/ void (*autosize)(WindowPtr ww), Boolean (*func)(WindowPtr wp))
{
	short maxWindowsPerRow, maxWindowsPerColumn;
	short screenWidth, screenHeight;
	TileStackWindowData data;
	
	//count windows
	data.currentWindow = data.numberOfWindows = 0; 
	ForEachWindowDo(FEWCountWindows,(Ptr)&data.numberOfWindows,func);
	
	if (data.numberOfWindows > 0) {
		data.screenRect = qd.screenBits.bounds;
		data.screenRect.top += GetMBarHeight();
		data.screenRect.top += kGapBetweenWindows; // MDT - right & bottom are automatically
		data.screenRect.left += kGapBetweenWindows;// adjusted in our FEW routine.
		
		screenWidth = data.screenRect.right - data.screenRect.left;
		screenHeight = data.screenRect.bottom - data.screenRect.top;
		
		data.numberOfRows = horv; // we re-interperet this as horv.
		if (horv == kTileHorizontal) {
			data.width = screenWidth;
			data.height = screenHeight/data.numberOfWindows - kGapBetweenWindows;
		} else {
			data.width = screenWidth / data.numberOfWindows - kGapBetweenWindows;
			data.height = screenHeight;
		}	
		data.autosize = autosize;
		ForEachWindowDo(FEWTileWindows,(Ptr)&data,func);
	}
}

/* SetStdZoomRect
 */
void SetStdZoomRect (Rect *r)
{
	(*r) = qd.screenBits.bounds;
	r->top += GetMBarHeight() + 18; // 18 is the (approx) width of the WindowTitleBar
	InsetRect(r, kGapBetweenWindows, kGapBetweenWindows);
	r->bottom-= kGapBetweenWindows; // just to make sure
	r->right -= kGapBetweenWindows;
}

/* SetWindowBounds
 * Stolen from Scott Knaster
 */
void SetWindowBounds(WindowPtr w, Rect *r)
{
	Rect newBounds,oldBounds;
	short top, left, height, width;
	short topInset, leftInset, bottomInset, rightInset;
	RgnHandle contRgn,strucRgn;
	
	if (!r) return;
	if (!w) return;
	newBounds = *r;
	
	contRgn = ((WindowPeek)w)->contRgn;
	strucRgn = ((WindowPeek)w)->strucRgn;
	oldBounds = (*strucRgn)->rgnBBox;
	
	if (!EqualRect(&oldBounds, &newBounds)) {
		topInset =    (*contRgn)->rgnBBox.top    - (*strucRgn)->rgnBBox.top;
		leftInset =   (*contRgn)->rgnBBox.left   - (*strucRgn)->rgnBBox.left;
		bottomInset = (*strucRgn)->rgnBBox.bottom - (*contRgn)->rgnBBox.bottom;
		rightInset =  (*strucRgn)->rgnBBox.right  - (*contRgn)->rgnBBox.right;

		top = newBounds.top + topInset;
		left = newBounds.left + leftInset;
		height = newBounds.bottom - top - bottomInset;
		width = newBounds.right - left - rightInset;
		
		HideWindow(w);
		MoveWindow(w,left,top,false);
		SizeWindow(w,width,height,true);
		SelectWindow(w);
		ShowWindow(w);
	}
}

void ForEachWindowDo(void (*func)(WindowPtr wp, Ptr d), Ptr data, Boolean (*exclude)(WindowPtr wpp))
{
	WindowPtr w;
	short kind;
	Boolean exc;
	WindowListPtr head=nil,cur=nil;
	
	w = FrontWindow();
	
	while (w) {
		kind = ((WindowPeek)w)->windowKind;
		if (kind == userKind) {
			exc = false;
			if (exclude) exc = (*exclude)(w);
			if (!exc) {
				// add to our mini-list; push onto list like a stack.
				cur = (WindowListPtr) NewPtr(sizeof(WindowListNode));
				cur->w = w;
				cur->next=head;
				head = cur;
			} 
		}
		w = (WindowPtr) ((WindowPeek)w)->nextWindow;
	}
	
	//travel down our list.
	while (head) {
		cur = head;
		head = head->next; // pop
		(*func)(cur->w,data); //exec
		DisposePtr((Ptr)cur); //dispose
	}
}

void FEWTileWindows(WindowPtr w, Ptr d)
{
	Rect newBounds;
	short row,col;
	TileStackWindowData *data;
	
	if ((!w) || (!d)) return;
	data = (TileStackWindowData*)d;
	data->currentWindow++;
		
	if (data->numberOfRows == kTileHorizontal) {
		newBounds.left = data->screenRect.left;
		newBounds.right = newBounds.left + data->width - kGapBetweenWindows;
		newBounds.top = data->screenRect.top + ((data->currentWindow -1) * (data->height + kGapBetweenWindows));
		newBounds.bottom = newBounds.top + data->height;
	} else {
		newBounds.left = data->screenRect.left + ((data->currentWindow -1) * (data->width + kGapBetweenWindows));
		newBounds.right = newBounds.left + data->width;
		newBounds.top = data->screenRect.top;
		newBounds.bottom = newBounds.top + data->height - kGapBetweenWindows;
	}
	
	SetWindowBounds(w,&newBounds);
	if (data->autosize) (*(data->autosize))(w);
}

void FEWStackWindows(WindowPtr w, Ptr d)
{
	TileStackWindowData *data;
	Rect newBounds;
	short placeInDiagonal, diagonal;
	
	if ((!w) || (!d)) return;
	data = (TileStackWindowData *)d;
	data->currentWindow++;

	placeInDiagonal = (data->currentWindow - 1) % kWindowsPerDiagonal;
	diagonal = (data->currentWindow - 1) / kWindowsPerDiagonal;
	
	newBounds.left = data->screenRect.left + (data->currentWindow - 1) * kHorizontalStagger;
	newBounds.top  = data->screenRect.top + placeInDiagonal * kVerticalStagger;
	newBounds.right= newBounds.left + data->width - diagonal * (kWindowsPerDiagonal * kHorizontalStagger - 10);
	newBounds.bottom=newBounds.top  + data->height /*- placeInDiagonal * 10*/;

	SetWindowBounds(w,&newBounds);
	if (data->autosize) (*(data->autosize))(w);
}

void FEWCountWindows(WindowPtr w, Ptr data)
{
	short *count = (short*)data;
	
	*count += 1;
}

